searchNodes=[{"doc":"Prints Nebulex tasks and their information. mix nebulex","ref":"Mix.Tasks.Nebulex.html","title":"Mix.Tasks.Nebulex","type":"task"},{"doc":"Generates a new cache. The cache will be placed in the lib directory. Examples mix nebulex.gen.cache -c Custom.Cache -a Nebulex.Adapters.Local Command line options -c, --cache - the cache to generate -a, --adapter - the cache adapter to use (optional)","ref":"Mix.Tasks.Nebulex.Gen.Cache.html","title":"Mix.Tasks.Nebulex.Gen.Cache","type":"task"},{"doc":"Nebulex is split into 2 main components: Nebulex.Cache - caches are wrappers around the in-memory data store. Via the cache, we can insert, retrieve, update and delete entries from the in-memory data store. A cache needs an adapter to communicate to the in-memory data store. Nebulex.Object - this is the struct used by caches to handle data back and forth, entries in cache are stored and retrieved as objects, but this is totally handled by the cache adapter. In the following sections, we will provide an overview of those components and how they interact with each other. Feel free to access their respective module documentation for more specific examples, options and configuration. If you want to quickly check a sample application using Nebulex, please check the getting started guide. Caches Nebulex.Cache is the wrapper around the Cache. We can define a cache as follows: defmodule MyApp.MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end Where the configuration for the Cache must be in your application environment, usually defined in your config/config.exs: config :my_app, MyApp.MyCache, n_shards: 2, gc_interval: 3600 Each cache in Nebulex defines a start_link/1 function that needs to be invoked before using the cache. In general, this function is not called directly, but used as part of your application supervision tree. If your application was generated with a supervisor (by passing --sup to mix new) you will have a lib/my_app/application.ex file containing the application start callback that defines and starts your supervisor. You just need to edit the start/2 function to start the repo as a supervisor on your application&#39;s supervisor: def start(_type, _args) do children = [ {MyApp.Cache, []} ] opts = [strategy: :one_for_one, name: MyApp.Supervisor] Supervisor.start_link(children, opts) end You can use the cache from your application whatever you want. For example, let&#39;s suppose we want to handle user sessions in cache: defmodule MyApp.UserSession do alias MyApp.Repo alias MyApp.Cache alias MyApp.User def new(username, password) do case Repo.get(User, username) do %User{password: ^password} = user -&gt; sesion_token = username |&gt; generate_token(password) |&gt; Cache.set(user, ttl: 3600) {:ok, sesion_token} _ -&gt; {:error, :unauthorized} end end def get(token) do case Cache.get(token) do nil -&gt; {:error, :not_found} user -&gt; {:ok, user} end end ## Helpers defp generate_token(username, password) do # your code ... end end Object Nebulex.Object is the struct used by the caches to store and retrieve data. By default, when we perform an action (set, get, delete, etc.) on a cache entry, the cache returns only the value of that entry, but you can ask for return the whole object: iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;, return: :object) %Nebulex.Object{key: &quot;foo&quot;, expire_at: nil, value: &quot;bar&quot;, version: nil} iex&gt; MyCache.get(&quot;foo&quot;, return: :object) %Nebulex.Object{key: &quot;foo&quot;, expire_at: nil, value: &quot;bar&quot;, version: nil} See Nebulex.Object to learn more about it. Object Versioning (Optimistic Offline Locks) It is also possible to set a version (or ETag) on the object to be cached; in the case you want to implement something like &quot;Optimistic Locking&quot; through a version property on cached objects. Supposing we have a &quot;serial&quot; version generator: iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;, return: :object) %Nebulex.Object{key: &quot;foo&quot;, expire_at: nil, value: &quot;bar&quot;, version: 1} iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;, version: 1, return: :object) %Nebulex.Object{key: &quot;foo&quot;, expire_at: nil, value: &quot;bar&quot;, version: 2} iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;, version: 2, return: :object) %Nebulex.Object{key: &quot;foo&quot;, expire_at: nil, value: &quot;bar&quot;, version: 3}","ref":"Nebulex.html","title":"Nebulex","type":"module"},{"doc":"This module specifies the adapter API that a Cache adapter is required to implement.","ref":"Nebulex.Adapter.html","title":"Nebulex.Adapter","type":"behaviour"},{"doc":"The callback invoked in case the adapter needs to inject code.","ref":"Nebulex.Adapter.html#c:__before_compile__/1","title":"Nebulex.Adapter.__before_compile__/1","type":"macrocallback"},{"doc":"Deletes a single object from cache. See Nebulex.Cache.delete/2.","ref":"Nebulex.Adapter.html#c:delete/3","title":"Nebulex.Adapter.delete/3","type":"callback"},{"doc":"Returns the expiry timestamp for the given key, if the timeout ttl (in seconds) is successfully updated. If key doesn&#39;t exist, nil is returned. See Nebulex.Cache.expire/2.","ref":"Nebulex.Adapter.html#c:expire/3","title":"Nebulex.Adapter.expire/3","type":"callback"},{"doc":"Flushes the cache. See Nebulex.Cache.flush/0.","ref":"Nebulex.Adapter.html#c:flush/1","title":"Nebulex.Adapter.flush/1","type":"callback"},{"doc":"Retrieves a single object from cache. See Nebulex.Cache.get/2.","ref":"Nebulex.Adapter.html#c:get/3","title":"Nebulex.Adapter.get/3","type":"callback"},{"doc":"Returns a map with the objects for all specified keys. For every key that does not hold a value or does not exist, that key is simply ignored. Because of this, the operation never fails. See Nebulex.Cache.get_many/2.","ref":"Nebulex.Adapter.html#c:get_many/3","title":"Nebulex.Adapter.get_many/3","type":"callback"},{"doc":"Returns whether the given key exists in cache. See Nebulex.Cache.has_key?/1.","ref":"Nebulex.Adapter.html#c:has_key?/2","title":"Nebulex.Adapter.has_key?/2","type":"callback"},{"doc":"Initializes the adapter supervision tree by returning the children","ref":"Nebulex.Adapter.html#c:init/1","title":"Nebulex.Adapter.init/1","type":"callback"},{"doc":"Returns the information associated with attr for the given key, or returns nil if key doesn&#39;t exist. See Nebulex.Cache.object_info/2.","ref":"Nebulex.Adapter.html#c:object_info/3","title":"Nebulex.Adapter.object_info/3","type":"callback"},{"doc":"Sets the given object under key into the cache. If the object already exists, it is overwritten. Any previous time to live associated with the key is discarded on successful set operation. Returns true if an object with key key is found and successfully inserted, otherwise false. Options Besides the &quot;Shared options&quot; section in Nebulex.Cache documentation, it accepts: :action - It may be one of :add, :replace, :set (the default). See the &quot;Actions&quot; section for more information. Actions The :action option supports the following values: :add - Only set the key if it does not already exist. If it does, false is returned. :replace - Alters the object stored under key, but only if the object already exists into the cache. :set - Set key to hold the given object (default). See Nebulex.Cache.set/3, Nebulex.Cache.add/3, Nebulex.Cache.replace/3.","ref":"Nebulex.Adapter.html#c:set/3","title":"Nebulex.Adapter.set/3","type":"callback"},{"doc":"Sets the given objects, replacing existing ones, just as regular set. Returns :ok if the all objects were successfully set, otherwise {:error, failed_keys}, where failed_keys contains the keys that could not be set. Ideally, this operation should be atomic, so all given keys are set at once. But it depends purely on the adapter&#39;s implementation and the backend used internally by the adapter. Hence, it is recommended to checkout the adapter&#39;s documentation. See Nebulex.Cache.set_many/2.","ref":"Nebulex.Adapter.html#c:set_many/3","title":"Nebulex.Adapter.set_many/3","type":"callback"},{"doc":"Returns the total number of cached entries. See Nebulex.Cache.size/0.","ref":"Nebulex.Adapter.html#c:size/1","title":"Nebulex.Adapter.size/1","type":"callback"},{"doc":"Returns and removes the object with key key in the cache. See Nebulex.Cache.take/2.","ref":"Nebulex.Adapter.html#c:take/3","title":"Nebulex.Adapter.take/3","type":"callback"},{"doc":"Updates (increment or decrement) the counter mapped to the given key. See Nebulex.Cache.update_counter/3.","ref":"Nebulex.Adapter.html#c:update_counter/4","title":"Nebulex.Adapter.update_counter/4","type":"callback"},{"doc":"","ref":"Nebulex.Adapter.html#t:cache/0","title":"Nebulex.Adapter.cache/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:key/0","title":"Nebulex.Adapter.key/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:object/0","title":"Nebulex.Adapter.object/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:opts/0","title":"Nebulex.Adapter.opts/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.html#t:t/0","title":"Nebulex.Adapter.t/0","type":"type"},{"doc":"This behaviour provides a callback to compute the hash slot for a specific key based on the number of slots/nodes. The purpose of this module is to allow users to implement a custom hash function to distribute keys. It can be used to select the node/slot where a specific key is supposed to be. It is highly recommended to use a Consistent Hashing algorithm. Example defmodule MyApp.HashSlot do @behaviour Nebulex.Adapter.HashSlot def keyslot(key, range) do key |&gt; :erlang.phash2() |&gt; :jchash.compute(range) end end This example uses Jumping Consistent Hash.","ref":"Nebulex.Adapter.HashSlot.html","title":"Nebulex.Adapter.HashSlot","type":"behaviour"},{"doc":"Computes the hash for a specific key based on the number of slots given by range. Returns a slot within the range 0..range-1. Example iex&gt; MyHash.keyslot(&quot;mykey&quot;, 10) 2","ref":"Nebulex.Adapter.HashSlot.html#c:keyslot/2","title":"Nebulex.Adapter.HashSlot.keyslot/2","type":"callback"},{"doc":"Specifies the adapter persistence API. Default implementation This module provides a default implementation that uses File and Stream under-the-hood. For dumping a cache to a file, the entries are streamed from the cache and written in chunks (one chunk per line), and each chunk contains N number of entries. For loading the entries from a file, the file is read and streamed line-by-line, so that the entries collected on each line are inserted in streaming fashion as well. The default implementation accepts the following options only for dump operation (there are not options for load): objects_per_line - The number of objects to be written per line in the file. Defaults to 10. compression - The compression level. The values are the same as :erlang.term_to_binary /2. Defaults to 6. See Nebulex.Cache.dump/2 and Nebulex.Cache.load/2 for more information.","ref":"Nebulex.Adapter.Persistence.html","title":"Nebulex.Adapter.Persistence","type":"behaviour"},{"doc":"Dumps a cache to the given file path. Returns :ok if successful, or {:error, reason} if an error occurs. See Nebulex.Cache.dump/2.","ref":"Nebulex.Adapter.Persistence.html#c:dump/3","title":"Nebulex.Adapter.Persistence.dump/3","type":"callback"},{"doc":"Loads a dumped cache from the given path. Returns :ok if successful, or {:error, reason} if an error occurs. See Nebulex.Cache.load/2.","ref":"Nebulex.Adapter.Persistence.html#c:load/3","title":"Nebulex.Adapter.Persistence.load/3","type":"callback"},{"doc":"Specifies the query API required from adapters.","ref":"Nebulex.Adapter.Queryable.html","title":"Nebulex.Adapter.Queryable","type":"behaviour"},{"doc":"Fetches all entries from cache matching the given query. If the query is nil, it fetches all entries from cache; this is common for all adapters. However, the query could be any other value, that depends entirely on the adapter&#39;s implementation. Therefore, it is recommended to check out adapters&#39; documentation. For instance, the built-in Nebulex.Adapters.Local adapter supports :ets.match_spec() as query. May raise Nebulex.QueryError if query validation fails. See Nebulex.Cache.all/2.","ref":"Nebulex.Adapter.Queryable.html#c:all/3","title":"Nebulex.Adapter.Queryable.all/3","type":"callback"},{"doc":"Streams the given query. It returns a stream of values. May raise Nebulex.QueryError if query validation fails. See Nebulex.Cache.stream/2.","ref":"Nebulex.Adapter.Queryable.html#c:stream/3","title":"Nebulex.Adapter.Queryable.stream/3","type":"callback"},{"doc":"","ref":"Nebulex.Adapter.Queryable.html#t:cache/0","title":"Nebulex.Adapter.Queryable.cache/0","type":"type"},{"doc":"","ref":"Nebulex.Adapter.Queryable.html#t:opts/0","title":"Nebulex.Adapter.Queryable.opts/0","type":"type"},{"doc":"Specifies the adapter transactions API. Default implementation This module also provides a default implementation which uses the Erlang library :global. This implementation accepts the following options: :keys - The list of the keys that will be locked. Since the lock id is generated based on the key, if this option is not set, a fixed/constant lock id is used to perform the transaction, then all further transactions (without this option set) are serialized and the performance is affected significantly. For that reason it is recommended to pass the list of keys involved in the transaction. :nodes - The list of the nodes where the lock will be set, or on all nodes if none are specified. :retries - If the key has been locked by other process already, and :retries is not equal to 0, the process sleeps for a while and tries to execute the action later. When :retries attempts have been made, an exception is raised. If :retries is :infinity (the default), the function will eventually be executed (unless the lock is never released). Let&#39;s see an example: MyCache.transaction fn -&gt; counter = MyCache.get(:counter) MyCache.set(:counter, counter + 1) end Locking only the involved key (recommended): MyCache.transaction fn -&gt; counter = MyCache.get(:counter) MyCache.set(:counter, counter + 1) end, keys: [:counter] MyCache.transaction fn -&gt; alice = MyCache.get(:alice) bob = MyCache.get(:bob) MyCache.set(:alice, %{alice | balance: alice.balance + 100}) MyCache.set(:bob, %{bob | balance: bob.balance + 100}) end, keys: [:alice, :bob]","ref":"Nebulex.Adapter.Transaction.html","title":"Nebulex.Adapter.Transaction","type":"behaviour"},{"doc":"Returns true if the given process is inside a transaction. See Nebulex.Cache.in_transaction?/0.","ref":"Nebulex.Adapter.Transaction.html#c:in_transaction?/1","title":"Nebulex.Adapter.Transaction.in_transaction?/1","type":"callback"},{"doc":"Runs the given function inside a transaction. A successful transaction returns the value returned by the function. See Nebulex.Cache.transaction/2.","ref":"Nebulex.Adapter.Transaction.html#c:transaction/3","title":"Nebulex.Adapter.Transaction.transaction/3","type":"callback"},{"doc":"Adapter module for Local Generational Cache. It uses Shards as in-memory backend (ETS tables are used internally). Features Support for generational cache – inspired by epocxy. Support for Sharding – handled by :shards. Support for garbage collection via Nebulex.Adapters.Local.Generation Support for transactions via Erlang global name registration facility Options These options can be set through the config file: :n_generations - Max number of Cache generations, defaults to 2. :n_shards - The number of partitions for each Cache generation table. Defaults to System.schedulers_online(). See :shards.new/2. :read_concurrency - Indicates whether the tables that :shards creates uses :read_concurrency or not (default: true). See :ets.new/2. :write_concurrency - Indicates whether the tables that :shards creates uses :write_concurrency or not (default: true). See :ets.new/2. :compressed - If this option is present, the table data is stored in a more compact format to consume less memory. However, it will make table operations slower (default: false). See :ets.new/2. :gc_interval - Interval time in seconds to garbage collection to run, delete the oldest generation and create a new one. If this option is not set, garbage collection is never executed, so new generations must be created explicitly, e.g.: MyCache.new_generation([]). :allocated_memory - Max size in bytes allocated for a cache generation. If this option is set and the configured value is reached, a new generation is created so the oldest is deleted and force releasing memory space. If it is not set (nil), the cleanup check to release memory is not performed (the default). :gc_cleanup_interval - The number of writes needed to run the cleanup check. Once this value is reached and only if allocated_memory option is set, the cleanup check is performed. Defaults to 10, so after 10 write operations the cleanup check is performed. Example Nebulex.Cache is the wrapper around the cache. We can define a local cache as follows: defmodule MyApp.LocalCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end Where the configuration for the cache must be in your application environment, usually defined in your config/config.exs: config :my_app, MyApp.LocalCache, n_shards: 2, gc_interval: 3600, write_concurrency: true For more information about the usage, check out Nebulex.Cache. Extended API This adapter provides some additional functions to the Nebulex.Cache API. Most of these functions are used internally by the adapter, but there is one function which is indeed provided to be used from the Cache API, and it is the function to create new generations: new_generation/1. MyApp.LocalCache.new_generation Other additional function that might be useful is: __metadata__, which is used to retrieve the Cache Metadata. MyApp.LocalCache.__metadata__ The rest of the functions as we mentioned before, are for internal use. Queryable API The adapter supports as query parameter the following values: query - nil | :all_unexpired | :all_expired | :ets.match_spec() Internally, an entry is represented by the tuple {key, val, vsn, exp}, which means the match pattern within the :ets.match_spec() must be something like {:&quot;$1&quot;, :&quot;$2&quot;, :&quot;$3&quot;, :&quot;$4&quot;}. In order to make query building easier, you can use Ex2ms library. Examples # built-in queries MyCache.all() MyCache.all(:all_unexpired) MyCache.all(:all_expired) # using a custom match spec (all values &gt; 10) spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 10}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] MyCache.all(spec) # using Ex2ms import Ex2ms spec = fun do {key, value, _version, _expire_at} when value &gt; 10 -&gt; {key, value} end MyCache.all(spec) The :return option applies only for built-in queries, such as: nil | :all_unexpired | :all_expired, if you are using a custom :ets.match_spec(), the return value depends on it. The same applies to stream function.","ref":"Nebulex.Adapters.Local.html","title":"Nebulex.Adapters.Local","type":"module"},{"doc":"Generations Handler. This GenServer acts as garbage collector, everytime it runs, a new cache generation is created a the oldest one is deleted. The only way to create new generations is through this module (this server is the metadata owner) calling new/2 function. When a Cache is created, a generations handler associated to that Cache is started at the same time, therefore, this server MUST NOT be started directly. Options These options are configured via the built-in local adapter (Nebulex.Adapters.Local): :gc_interval - Interval time in seconds to garbage collection to run, delete the oldest generation and create a new one. If this option is not set, garbage collection is never executed, so new generations must be created explicitly, e.g.: new(cache, []). :allocated_memory - Max size in bytes allocated for a cache generation. If this option is set and the configured value is reached, a new generation is created so the oldest is deleted and force releasing memory space. If it is not set (nil), the cleanup check to release memory is not performed (the default). :gc_cleanup_interval - The number of writes needed to run the cleanup check. Once this value is reached and only if allocated_memory option is set, the cleanup check is performed. Defaults to 10, so after 10 write operations the cleanup check is performed.","ref":"Nebulex.Adapters.Local.Generation.html","title":"Nebulex.Adapters.Local.Generation","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor.","ref":"Nebulex.Adapters.Local.Generation.html#child_spec/1","title":"Nebulex.Adapters.Local.Generation.child_spec/1","type":"function"},{"doc":"Triggers the cleanup process to check whether or not the max generation size has been reached. If so, a new generation is pushed in order to release memory and keep it within the configured limit. Example Nebulex.Adapters.Local.Generation.cleanup(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#cleanup/1","title":"Nebulex.Adapters.Local.Generation.cleanup/1","type":"function"},{"doc":"Flushes the cache (including all its generations). Example Nebulex.Adapters.Local.Generation.flush(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#flush/1","title":"Nebulex.Adapters.Local.Generation.flush/1","type":"function"},{"doc":"Returns the GenServer state (mostly for testing purposes). Example Nebulex.Adapters.Local.Generation.get_state(MyCache)","ref":"Nebulex.Adapters.Local.Generation.html#get_state/1","title":"Nebulex.Adapters.Local.Generation.get_state/1","type":"function"},{"doc":"Creates a new cache generation. Once the max number of generations is reached, when a new generation is created, the oldest one is deleted. Options :reset_timeout - Indicates if the poll frequency time-out should be reset or not (default: true). Example Nebulex.Adapters.Local.Generation.new(MyCache, reset_timeout: :false)","ref":"Nebulex.Adapters.Local.Generation.html#new/2","title":"Nebulex.Adapters.Local.Generation.new/2","type":"function"},{"doc":"Reallocates the block of memory that was previously allocated for the given cache with the new size. In other words, reallocates the max memory size for a cache generation. Example Nebulex.Adapters.Local.Generation.realloc(MyCache, 1_000_000)","ref":"Nebulex.Adapters.Local.Generation.html#realloc/2","title":"Nebulex.Adapters.Local.Generation.realloc/2","type":"function"},{"doc":"Starts the garbage collector for the build-in local cache adapter.","ref":"Nebulex.Adapters.Local.Generation.html#start_link/2","title":"Nebulex.Adapters.Local.Generation.start_link/2","type":"function"},{"doc":"Adapter module for Multi-level Cache. This is just a simple layer on top of local or distributed cache implementations that enables to have a cache hierarchy by levels. Multi-level caches generally operate by checking the fastest, level 1 (L1) cache first; if it hits, the adapter proceeds at high speed. If that first cache misses, the next fastest cache (level 2, L2) is checked, and so on, before accessing external memory (that can be handled by a :fallback function). For write functions, the &quot;Write Through&quot; policy is applied by default; this policy ensures that the data is stored safely as it is written throughout the hierarchy. However, it is possible to force the write operation in a specific level (although it is not recommended) via level option, where the value is a positive integer greater than 0. Options These options can be set through the config file: :cache_model - Specifies the cache model: :inclusive or :exclusive; defaults to :inclusive. In an inclusive cache, the same data can be present in all caches/levels. In an exclusive cache, data can be present in only one cache/level and a key cannot be found in the rest of caches at the same time. This option affects get operation only; if :cache_model is :inclusive, when the key is found in a level N, that entry is duplicated backwards (to all previous levels: 1..N-1). :levels - The list of caches where each cache corresponds to a level. The order of the caches in the list defines the the order of the levels as well, for example, the first cache in the list will be the L1 cache (level 1) and so on; the Nth elemnt will be the LN cache. This option is mandatory, if it is not set or empty, an exception will be raised. :fallback - Defines a fallback function when a key is not present in any cache level. Function is defined as: (key -&gt; value). The :fallback can be defined in two ways: at compile-time and at run-time. At compile-time, in the cache config, set the module that implements the function: config :my_app, MyApp.MultilevelCache, fallback: &amp;MyMapp.AnyModule/1 And at run-time, passing the function as an option within the opts argument (only valid for get function): MultilevelCache.get(&quot;foo&quot;, fallback: fn(key) -&gt; key * 2 end) Shared Options Some functions below accept the following options: :level - It may be an integer greater than 0 that specifies the cache level where the operation will take place. By default, the evaluation is performed throughout the whole cache hierarchy (all levels). Example Nebulex.Cache is the wrapper around the Cache. We can define the multi-level cache as follows: defmodule MyApp.MultilevelCache do use Nebulex.Cache, otp_app: :nebulex, adapter: Nebulex.Adapters.Multilevel defmodule L1 do use Nebulex.Cache, otp_app: :nebulex, adapter: Nebulex.Adapters.Local end defmodule L2 do use Nebulex.Cache, otp_app: :nebulex, adapter: Nebulex.Adapters.Partitioned end def fallback(_key) do # maybe fetch the data from Database nil end end defmodule MyApp.LocalCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end Where the configuration for the Cache must be in your application environment, usually defined in your config/config.exs: config :my_app, MyApp.MultilevelCache, levels: [ MyApp.MultilevelCache.L1, MyApp.MultilevelCache.L2 ], fallback: &amp;MyApp.MultilevelCache.fallback/1 config :my_app, MyApp.MultilevelCache.L1, n_shards: 2, gc_interval: 3600 config :my_app, MyApp.MultilevelCache.L2, primary: MyApp.LocalCache config :my_app, MyApp.LocalCache, n_shards: 2, gc_interval: 3600 Using the multilevel cache cache: # Retrieving data from cache MyCache.get(&quot;foo&quot;, fallback: fn(_key) -&gt; # Maybe fetch the key from database &quot;initial value&quot; end) # Entry is set in all cache levels MyCache.set(&quot;foo&quot;, &quot;bar&quot;) # Entry is set at second cache level MyCache.set(&quot;foo&quot;, &quot;bar&quot;, level: 2) # Entry is deleted from all cache levels MyCache.delete(&quot;foo&quot;) # Entry is deleted from second cache level MyCache.delete(&quot;foo&quot;, level: 2) Extended API This adapter provides some additional functions to the Nebulex.Cache API. __levels__ This function returns the configured level list. MyCache.__levels__ __model__ This function returns the multi-level cache model. MyCache.__model__ __fallback__ This function returns the default fallback function. MyCache.__fallback__ Limitations Because this adapter reuses other existing/configured adapters, it inherits all their limitations too. Therefore, it is highly recommended to check the documentation of the used adapters.","ref":"Nebulex.Adapters.Multilevel.html","title":"Nebulex.Adapters.Multilevel","type":"module"},{"doc":"Built-in adapter for partitioned cache topology. A partitioned cache is a clustered, fault-tolerant cache that has linear scalability. Data is partitioned among all the machines of the cluster. For fault-tolerance, partitioned caches can be configured to keep each piece of data on one or more unique machines within a cluster. This adapter in particular hasn&#39;t fault-tolerance built-in, each piece of data is kept in a single node/machine (sharding), therefore, if a node fails, the data kept by this node won&#39;t be available for the rest of the cluster. This adapter depends on a local cache adapter (primary storage), it adds a thin layer on top of it in order to distribute requests across a group of nodes, where is supposed the local cache is running already. PG2 is used under-the-hood by the adapter to manage the cluster nodes. When the partitioned cache is started in a node, it creates a PG2 group and joins it (the cache supervisor PID is joined to the group). Then, when a function is invoked, the adapter picks a node from the node list (using the PG2 group members), and then the function is executed on that node. In the same way, when the supervisor process of the partitioned cache dies, the PID of that process is automatically removed from the PG2 group; this is why it&#39;s recommended to use a consistent hashing algorithm for the node selector. Features Support for partitioned topology (Sharding Distribution Model) Support for transactions via Erlang global name registration facility Configurable hash-slot module to compute the node Options These options can be set through the config file: :primary - The module for the primary storage. The value must be a valid local cache adapter so that the partitioned adapter can store the data in there. For example, you can set the Nebulex.Adapters.Local as value, unless you want to provide another one. :hash_slot - The module that implements Nebulex.Adapter.HashSlot behaviour. Runtime options These options apply to all adapter&#39;s functions. :timeout - The time-out value in milliseconds for the command that will be executed. If the timeout is exceeded, then the current process will exit. This adapter uses Task.await/2 internally, therefore, check the function documentation to learn more about it. For commands like set_many and get_many, if the timeout is exceeded, the task is shutted down but the current process doesn&#39;t exit, only the result associated to that task is just skipped in the reduce phase. task_supervisor_opts - Defines the options passed to Task.Supervisor.start_link/1 when the adapter is initialized. Example Nebulex.Cache is the wrapper around the cache. We can define the partitioned cache as follows: defmodule MyApp.PartitionedCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Partitioned defmodule Primary do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end end Where the configuration for the cache must be in your application environment, usually defined in your config/config.exs: config :my_app, MyApp.PartitionedCache.Primary, n_shards: 2, gc_interval: 3600 config :my_app, MyApp.PartitionedCache, primary: MyApp.PartitionedCache.Primary For more information about the usage, check out Nebulex.Cache. Extended API This adapter provides some additional functions to the Nebulex.Cache API. __primary__ Returns the local cache adapter (the local backend). __task_sup__ Returns the task supervisor module that manages RPC calls. __nodes__ Returns the nodes that belongs to the caller Cache. get_node/1 This function invokes c:Nebulex.Adapter.NodeSelector.get_node/2 internally. MyCache.get_node(&quot;mykey&quot;) Limitations This adapter has a limitation for two functions: get_and_update/4 and update/5. They both have a parameter that is the anonymous function, and the anonymous function is compiled into the module where it is created, which means it necessarily doesn&#39;t exists on remote nodes. To ensure they work as expected, you must provide functions from modules existing in all nodes of the group.","ref":"Nebulex.Adapters.Partitioned.html","title":"Nebulex.Adapters.Partitioned","type":"module"},{"doc":"Helper to perform stream/3 locally.","ref":"Nebulex.Adapters.Partitioned.html#eval_local_stream/3","title":"Nebulex.Adapters.Partitioned.eval_local_stream/3","type":"function"},{"doc":"Built-in adapter for replicated cache topology. The replicated cache excels in its ability to handle data replication, concurrency control and failover in a cluster, all while delivering in-memory data access speeds. A clustered replicated cache is exactly what it says it is: a cache that replicates its data to all cluster nodes. There are several challenges to building a reliably replicated cache. The first is how to get it to scale and perform well. Updates to the cache have to be sent to all cluster nodes, and all cluster nodes have to end up with the same data, even if multiple updates to the same piece of data occur at the same time. Also, if a cluster node requests a lock, ideally it should not have to get all cluster nodes to agree on the lock or at least do it in a very efficient way (:global is used for this), otherwise it will scale extremely poorly; yet in the case of a cluster node failure, all of the data and lock information must be kept safely. The best part of a replicated cache is its access speed. Since the data is replicated to each cluster node, it is available for use without any waiting. This is referred to as &quot;zero latency access,&quot; and is perfect for situations in which an application requires the highest possible speed in its data access. However, there are some limitations: Cost Per Update - Updating a replicated cache requires pushing the new version of the data to all other cluster members, which will limit scalability if there is a high frequency of updates per member. Cost Per Entry - The data is replicated to every cluster member, so Memory Heap space is used on each member, which will impact performance for large caches. Based on &quot;Distributed Caching Essential Lessons&quot; by Cameron Purdy. Options These options can be set through the config file: :primary - The module for the primary storage. The value must be a valid local cache adapter so that the partitioned adapter can store the data in there. For example, you can set the Nebulex.Adapters.Local as value, unless you want to provide another one. Runtime options These options apply to all adapter&#39;s functions. :timeout - The time-out value in milliseconds for the command that will be executed. If the timeout is exceeded, then the current process will exit. This adapter uses Task.await/2 internally, therefore, check the function documentation to learn more about it. For commands like set_many and get_many, if the timeout is exceeded, the task is shutted down but the current process doesn&#39;t exit, only the result associated to that task is just skipped in the reduce phase. task_supervisor_opts - Defines the options passed to Task.Supervisor.start_link/1 when the adapter is initialized. Example Nebulex.Cache is the wrapper around the cache. We can define the local and replicated cache as follows: defmodule MyApp.ReplicatedCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Replicated defmodule Primary do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end end Where the configuration for the cache must be in your application environment, usually defined in your config/config.exs: config :my_app, MyApp.ReplicatedCache.Primary, n_shards: 2, gc_interval: 3600 config :my_app, MyApp.ReplicatedCache, primary: MyApp.ReplicatedCache.Primary For more information about the usage, check out Nebulex.Cache. Extended API This adapter provides some additional functions to the Nebulex.Cache API. __primary__ Returns the local cache adapter (the local backend). __task_sup__ Returns the task supervisor module that manages RPC calls. __nodes__ Returns the nodes that belongs to the caller Cache.","ref":"Nebulex.Adapters.Replicated.html","title":"Nebulex.Adapters.Replicated","type":"module"},{"doc":"Cache Main Interface. A Cache maps to an underlying implementation, controlled by the adapter. For example, Nebulex ships with a default adapter that implements a local generational cache. When used, the Cache expects the :otp_app and :adapter as options. The :otp_app should point to an OTP application that has the Cache configuration. For example, the Cache: defmodule MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local end Could be configured with: config :my_app, MyCache, version_generator: MyCache.VersionGenerator, stats: true, gc_interval: 3600 Most of the configuration that goes into the config is specific to the adapter, so check Nebulex.Adapters.Local documentation for more information. However, some configuration is shared across all adapters, they are: :version_generator - this option specifies the module that implements the Nebulex.Object.Version interface. This interface defines only one callback generate/1 that is invoked by the adapters to generate new object versions. If this option is not set, then the version is set to nil by default. :stats - a compile-time option that specifies if cache statistics is enabled or not (defaults to false). Pre/Post hooks The cache can also provide its own pre/post hooks implementation; see Nebulex.Hook behaviour. By default, pre/post hooks are empty lists, but you can override the functions of Nebulex.Hook behaviour. Additionally, it is possible to configure the mode how the hooks are evaluated (synchronous, asynchronous and pipeline). For more information about the usage, check out Nebulex.Hook. Shared options Almost all of the Cache operations below accept the following options: :return - Selects return type: :value | :key | :object. If :value (the default) is set, only object&#39;s value is returned. If :key set, only object&#39;s key is returned. If :object is set, Nebulex.Object.t() is returned. :version - The version of the object on which the operation will take place. The version can be any term (default: nil). :ttl - Time To Live (TTL) or expiration time in seconds for a key (default: :infinity). Such cases will be explicitly documented as well as any extra option. Extended API Some adapters might extend the API with additional functions, therefore, it is important to check out adapters&#39; documentation.","ref":"Nebulex.Cache.html","title":"Nebulex.Cache","type":"behaviour"},{"doc":"Returns the adapter tied to the cache.","ref":"Nebulex.Cache.html#c:__adapter__/0","title":"Nebulex.Cache.__adapter__/0","type":"callback"},{"doc":"Sets the given value under key into the cache, only if it does not already exist. If cache doesn&#39;t contain the given key, then {:ok, value} is returned. If cache contains key, :error is returned. Options See the &quot;Shared options&quot; section at the module documentation. For add operation, the option :version is ignored. Example iex&gt; MyCache.add(&quot;foo&quot;, &quot;bar&quot;) {ok, &quot;bar&quot;} iex&gt; MyCache.add(&quot;foo&quot;, &quot;bar&quot;) :error # if the value is nil, it is not stored (operation is skipped) iex&gt; MyCache.add(&quot;foo&quot;, nil) {ok, nil}","ref":"Nebulex.Cache.html#c:add/3","title":"Nebulex.Cache.add/3","type":"callback"},{"doc":"Similar to add/3 but raises Nebulex.KeyAlreadyExistsError if the key already exists. Options See the &quot;Shared options&quot; section at the module documentation. Example MyCache.add!(&quot;foo&quot;, &quot;bar&quot;)","ref":"Nebulex.Cache.html#c:add!/3","title":"Nebulex.Cache.add!/3","type":"callback"},{"doc":"When the key already exists, the cached value is replaced by value, otherwise the object is created at fisrt time. Returns the replaced or created value when the function is completed successfully. Because this function is a combination of replace/3 and add!/3 functions, there might be a race condition between them, in that case the last operation add!/3 fails with Nebulex.KeyAlreadyExistsError. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - Same as callback set/3. Example MyCache.add_or_replace!(&quot;foo&quot;, &quot;bar&quot;)","ref":"Nebulex.Cache.html#c:add_or_replace!/3","title":"Nebulex.Cache.add_or_replace!/3","type":"callback"},{"doc":"Fetches all entries from cache matching the given query. If the query is nil, it fetches all entries from cache; this is common for all adapters. However, the query could be any other value, which depends entirely on the adapter&#39;s implementation; check out the &quot;Query&quot; section below. May raise Nebulex.QueryError if query validation fails. Options See the &quot;Shared options&quot; section at the module documentation. Note that for this function, option :version is ignored. Example # set some entries iex&gt; :ok = Enum.each(1..5, &amp;MyCache.set(&amp;1, &amp;1 * 2)) # fetch all (with default params) iex&gt; MyCache.all() [1, 2, 3, 4, 5] # fetch all entries and return values iex&gt; MyCache.all(nil, return: :value) [2, 4, 6, 8, 10] # fetch all entries and return objects iex&gt; [%Nebulex.Object{} | _] = MyCache.all(nil, return: :object) # fetch all entries that match with the given query # assuming we are using Nebulex.Adapters.Local adapter iex&gt; query = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 5}], [:&quot;$1&quot;]}] iex&gt; MyCache.all(query) [3, 4, 5] Query Query spec is defined by the adapter, hence, it is recommended to check out adapters documentation. For instance, the built-in Nebulex.Adapters.Local adapter supports nil | :all_unexpired | :all_expired | :ets.match_spec() as query value. Examples # additional built-in queries for Nebulex.Adapters.Local adapter iex&gt; all_unexpired = MyCache.all(:all_unexpired) iex&gt; all_expired = MyCache.all(:all_expired) # if we are using Nebulex.Adapters.Local adapter, the stored entry # is a tuple {key, value, version, expire_at}, then the match spec # could be something like: iex&gt; spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 5}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] iex&gt; MyCache.all(spec) [{3, 6}, {4, 8}, {5, 10}] # the same previous query but using Ex2ms iex&gt; import Ex2ms Ex2ms iex&gt; spec = ...&gt; fun do ...&gt; {key, value, _, _} when value &gt; 5 -&gt; {key, value} ...&gt; end iex&gt; MyCache.all(spec) [{3, 6}, {4, 8}, {5, 10}] To learn more, check out adapters&#39; documentation.","ref":"Nebulex.Cache.html#c:all/2","title":"Nebulex.Cache.all/2","type":"callback"},{"doc":"Returns the adapter configuration stored in the :otp_app environment. If the c:init/2 callback is implemented in the cache, it will be invoked.","ref":"Nebulex.Cache.html#c:config/0","title":"Nebulex.Cache.config/0","type":"callback"},{"doc":"Deletes the entry in cache for a specific key. If the key does not exist, returns the result according to the :return option (for delete defaults to :key) but the cache is not altered. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - It may be one of :raise (the default), :nothing, :override. See the &quot;OnConflict&quot; section for more information. Note that for this function :return option hasn&#39;t any effect since it always returns the key either success or not. Example iex&gt; MyCache.set(:a, 1) 1 iex&gt; MyCache.delete(:a) :a iex&gt; MyCache.get(:a) nil iex&gt; MyCache.delete(:non_existent_key) :non_existent_key OnConflict The :on_conflict option supports the following values: :raise - raises if there is a conflicting key :nothing - ignores the error in case of conflicts :override - the command is executed ignoring the conflict, then the value on the existing key is deleted Examples # Set a value iex&gt; MyCache.set(:a, 1) 1 # Delete with an invalid version but do nothing on conflicts. # Keep in mind that, although this returns successfully, the returned # `key` does not reflect the data in the cache. For instance, in case # of &quot;on_conflict: :nothing&quot;, the returned `key` isn&#39;t deleted. iex&gt; MyCache.delete(:a, version: :invalid, on_conflict: :nothing) :a iex&gt; MyCache.get(:a) 1 # Delete with the same invalid version but force to delete the current # value on conflicts (if the entry exists). iex&gt; MyCache.delete(:a, version: :invalid, on_conflict: :override) :a iex&gt; MyCache.get(:a) nil","ref":"Nebulex.Cache.html#c:delete/2","title":"Nebulex.Cache.delete/2","type":"callback"},{"doc":"Dumps a cache to the given file path. Returns :ok if successful, or {:error, reason} if an error occurs. Options This operation relies entirely on the adapter implementation, which means the options depend on each of them. For that reason, it is recommended to check the documentation of the adapter to be used. The built-in adapters inherit the default implementation from Nebulex.Adapter.Persistence, so check out the available options there. Examples # set some entries iex&gt; entries = for x &lt;- 1..10, into: %{}, do: {x, x} iex&gt; MyCache.set_many(entries) :ok # dump cache to a file iex&gt; MyCache.dump(&quot;my_cache&quot;) :ok To learn more, check out adapters&#39; documentation.","ref":"Nebulex.Cache.html#c:dump/2","title":"Nebulex.Cache.dump/2","type":"callback"},{"doc":"Returns the expiry timestamp for the given key, if the timeout ttl (in seconds) is successfully updated. If key doesn&#39;t exist, nil is returned. Examples iex&gt; MyCache.set(:a, 1) 1 iex&gt; MyCache.expire(:a, 5) 1540004049 iex&gt; MyCache.expire(:a, :infinity) :infinity iex&gt; MyCache.ttl(:b, 5) nil","ref":"Nebulex.Cache.html#c:expire/2","title":"Nebulex.Cache.expire/2","type":"callback"},{"doc":"Flushes the cache. Examples iex&gt; :ok = Enum.each(1..5, &amp;MyCache.set(&amp;1, &amp;1)) iex&gt; MyCache.flush() :ok iex&gt; Enum.each(1..5, fn x -&gt; ...&gt; nil = MyCache.get(x) ...&gt; end)","ref":"Nebulex.Cache.html#c:flush/0","title":"Nebulex.Cache.flush/0","type":"callback"},{"doc":"Gets a value or object from Cache where the key matches the given key. Returns nil if no result was found. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - It may be one of :raise (the default), :nothing, :override. See the &quot;OnConflict&quot; section for more information. Example iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;) &quot;bar&quot; iex&gt; MyCache.get(&quot;foo&quot;) &quot;bar&quot; iex&gt; MyCache.get(&quot;foo&quot;, return: :object) %Nebulex.Object{key: &quot;foo&quot;, value: &quot;bar&quot;, version: nil, expire_at: nil} iex&gt; MyCache.get(:non_existent_key) nil OnConflict The :on_conflict option supports the following values: :raise - raises if there is a conflicting key :nothing - ignores the error in case of conflicts :override - same effect as :nothing Examples # Set a value iex&gt; MyCache.set(:a, 1) 1 # Gets with an invalid version but do nothing on conflicts. # Keep in mind that, although this returns successfully, the returned # struct does not reflect the data in the Cache. For instance, in case # of &quot;on_conflict: :nothing&quot;, it returns the latest version of the # cached object. iex&gt; %Nebulex.Object{value: 1} = ...&gt; MyCache.get(:a, return: :object, version: :invalid, on_conflict: :nothing) iex&gt; MyCache.get(:a) 1","ref":"Nebulex.Cache.html#c:get/2","title":"Nebulex.Cache.get/2","type":"callback"},{"doc":"Similar to get/2 but raises KeyError if key is not found. Options See the &quot;Shared options&quot; section at the module documentation. Example MyCache.get!(:a)","ref":"Nebulex.Cache.html#c:get!/2","title":"Nebulex.Cache.get!/2","type":"callback"},{"doc":"Gets the object/value from key and updates it, all in one pass. fun is called with the current cached value under key (or nil if key hasn&#39;t been cached) and must return a two-element tuple: the &quot;get&quot; value (the retrieved value, which can be operated on before being returned) and the new value to be stored under key. fun may also return :pop, which means the current value shall be removed from Cache and returned. The returned value is a tuple with the &quot;get&quot; value returned by fun and the new updated value under key. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - Same as callback set/3. Examples # update nonexistent key iex&gt; MyCache.get_and_update(:a, fn current_value -&gt; ...&gt; {current_value, &quot;value!&quot;} ...&gt; end) {nil, &quot;value!&quot;} # update existing key iex&gt; MyCache.get_and_update(:a, fn current_value -&gt; ...&gt; {current_value, &quot;new value!&quot;} ...&gt; end) {&quot;value!&quot;, &quot;new value!&quot;} # pop/remove value if exist iex&gt; MyCache.get_and_update(:a, fn _ -&gt; :pop end) {&quot;new value!&quot;, nil} # pop/remove nonexistent key iex&gt; MyCache.get_and_update(:b, fn _ -&gt; :pop end) {nil, nil} # update existing key but returning the object iex&gt; {&quot;hello&quot;, %Object{key: :a, value: &quot;hello world&quot;}} = ...&gt; :a ...&gt; |&gt; MyCache.set!(&quot;hello&quot;, return: :key) ...&gt; |&gt; MyCache.get_and_update(fn current_value -&gt; ...&gt; {current_value, &quot;hello world&quot;} ...&gt; end, return: :object)","ref":"Nebulex.Cache.html#c:get_and_update/3","title":"Nebulex.Cache.get_and_update/3","type":"callback"},{"doc":"Returns a map with the values or objects (check :return option) for all specified keys. For every key that does not hold a value or does not exist, that key is simply ignored. Because of this, the operation never fails. Options See the &quot;Shared options&quot; section at the module documentation. Note that for this function, option :version is ignored. Example iex&gt; MyCache.set_many([a: 1, c: 3]) :ok iex&gt; MyCache.get_many([:a, :b, :c]) %{a: 1, c: 3}","ref":"Nebulex.Cache.html#c:get_many/2","title":"Nebulex.Cache.get_many/2","type":"callback"},{"doc":"Returns whether the given key exists in cache. Examples iex&gt; MyCache.set(:a, 1) 1 iex&gt; MyCache.has_key?(:a) true iex&gt; MyCache.has_key?(:b) false","ref":"Nebulex.Cache.html#c:has_key?/1","title":"Nebulex.Cache.has_key?/1","type":"callback"},{"doc":"Returns true if the current process is inside a transaction. Examples MyCache.in_transaction? #=&gt; false MyCache.transaction(fn -&gt; MyCache.in_transaction? #=&gt; true end)","ref":"Nebulex.Cache.html#c:in_transaction?/0","title":"Nebulex.Cache.in_transaction?/0","type":"callback"},{"doc":"A callback executed when the cache starts or when configuration is read.","ref":"Nebulex.Cache.html#c:init/1","title":"Nebulex.Cache.init/1","type":"callback"},{"doc":"Loads a dumped cache from the given path. Returns :ok if successful, or {:error, reason} if an error occurs. Options Similar to dump/2, this operation relies entirely on the adapter implementation, therefore, it is recommended to check the documentation of the adapter to be used. Similarly, the built-in adapters inherit the default implementation from Nebulex.Adapter.Persistence, so check out the available options there. Examples # set some entries iex&gt; entries = for x &lt;- 1..10, into: %{}, do: {x, x} iex&gt; MyCache.set_many(entries) :ok # dump cache to a file iex&gt; MyCache.dump(&quot;my_cache&quot;) :ok # load the cache from a file iex&gt; MyCache.load(&quot;my_cache&quot;) :ok To learn more, check out adapters&#39; documentation.","ref":"Nebulex.Cache.html#c:load/2","title":"Nebulex.Cache.load/2","type":"callback"},{"doc":"Returns the information associated with attr for the given key, or returns nil if key doesn&#39;t exist. If attr is not one of the allowed values, ArgumentError is raised. The following values are allowed for attr: :ttl - Returns the remaining time to live for the given key, if it has a timeout, otherwise, :infinity is returned. :version - Returns the current version for the given key. Examples iex&gt; MyCache.set(:a, 1, ttl: 5) 1 iex&gt; MyCache.set(:b, 2) 2 iex&gt; MyCache.object_info(:a, :ttl) 5 iex&gt; MyCache.ttl(:b) :infinity iex&gt; MyCache.object_info(:a, :version) nil","ref":"Nebulex.Cache.html#c:object_info/2","title":"Nebulex.Cache.object_info/2","type":"callback"},{"doc":"Alters the object stored under key, but only if the object already exists into the cache. If cache contains the given key, then {:ok, return} is returned. If cache doesn&#39;t contain key, :error is returned. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - Same as callback set/3. Example iex&gt; MyCache.replace(&quot;foo&quot;, &quot;bar&quot;) :error iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;) &quot;bar&quot; # update only current value iex&gt; MyCache.replace(&quot;foo&quot;, &quot;bar2&quot;) {:ok, &quot;bar2&quot;} # update current value and TTL iex&gt; MyCache.replace(&quot;foo&quot;, &quot;bar3&quot;, ttl: 10) {:ok, &quot;bar3&quot;}","ref":"Nebulex.Cache.html#c:replace/3","title":"Nebulex.Cache.replace/3","type":"callback"},{"doc":"Similar to replace/3 but raises KeyError if key is not found. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - Same as callback set/3. Example MyCache.replace!(&quot;foo&quot;, &quot;bar&quot;)","ref":"Nebulex.Cache.html#c:replace!/3","title":"Nebulex.Cache.replace!/3","type":"callback"},{"doc":"Sets the given value under key into the cache. If key already holds an object, it is overwritten. Any previous time to live associated with the key is discarded on successful set operation. Returns either the value, key or the object, depending on :return option, if the action is completed successfully. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - It may be one of :raise (the default), :nothing, :override. See the &quot;OnConflict&quot; section for more information. Example iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;) &quot;bar&quot; iex&gt; MyCache.set(&quot;foo&quot;, &quot;bar&quot;, ttl: 10, return: :object) %Nebulex.Object{key: &quot;foo&quot;, value: &quot;bar&quot;, version: nil, expire_at: 1540004049} # if the value is nil, then it is not stored (operation is skipped) iex&gt; MyCache.set(&quot;foo&quot;, nil) nil OnConflict The :on_conflict option supports the following values: :raise - raises if there is a conflicting key :nothing - ignores the error in case of conflicts :override - the command is executed ignoring the conflict, then the value on the existing key is replaced with the given value Examples iex&gt; %Nebulex.Object{version: v1} = MyCache.set(:a, 1, return: :object) iex&gt; MyCache.set(:a, 2, version: v1) 2 # Set with the same key and wrong version but do nothing on conflicts. # Keep in mind in case of &quot;on_conflict: :nothing&quot;, the returned object # is the current cached object, if there is one iex&gt; MyCache.set(:a, 3, version: v1, on_conflict: :nothing) 2 # Set with the same key and wrong version but replace the current # value on conflicts. iex&gt; MyCache.set(:a, 3, version: v1, on_conflict: :override) 3","ref":"Nebulex.Cache.html#c:set/3","title":"Nebulex.Cache.set/3","type":"callback"},{"doc":"Sets the given entries, replacing existing ones, just as regular set. Returns :ok if the all entries were successfully set, otherwise {:error, failed_keys}, where failed_keys contains the keys that could not be set. Ideally, this operation should be atomic, so all given keys are set at once. But it depends purely on the adapter&#39;s implementation and the backend used internally by the adapter. Hence, it is recommended to checkout the adapter&#39;s documentation. Options See the &quot;Shared options&quot; section at the module documentation. Note that for this function, option :version is ignored. Example iex&gt; MyCache.set_many(apples: 3, bananas: 1) :ok iex&gt; MyCache.set_many(%{&quot;apples&quot; =&gt; 1, &quot;bananas&quot; =&gt; 3}) :ok # set a custom list of objects iex&gt; MyCache.set_many([ ...&gt; %Nebulex.Object{key: :apples, value: 1}, ...&gt; %Nebulex.Object{key: :bananas, value: 2, expire_at: 5} ...&gt; ]) :ok # for some reason `:c` couldn&#39;t be set, so we got an error iex&gt; MyCache.set_many([a: 1, b: 2, c: 3], ttl: 1000) {:error, [:c]}","ref":"Nebulex.Cache.html#c:set_many/2","title":"Nebulex.Cache.set_many/2","type":"callback"},{"doc":"Returns the total number of cached entries. Examples iex&gt; :ok = Enum.each(1..10, &amp;MyCache.set(&amp;1, &amp;1)) iex&gt; MyCache.size() 10 iex&gt; :ok = Enum.each(1..5, &amp;MyCache.delete(&amp;1)) iex&gt; MyCache.size() 5","ref":"Nebulex.Cache.html#c:size/0","title":"Nebulex.Cache.size/0","type":"callback"},{"doc":"Starts a supervision and return {:ok, pid} or just :ok if nothing needs to be done. Returns {:error, {:already_started, pid}} if the cache is already started or {:error, term} in case anything else goes wrong. Options See the configuration in the moduledoc for options shared between adapters, for adapter-specific configuration see the adapter&#39;s documentation.","ref":"Nebulex.Cache.html#c:start_link/1","title":"Nebulex.Cache.start_link/1","type":"callback"},{"doc":"Shuts down the cache represented by the given pid.","ref":"Nebulex.Cache.html#c:stop/2","title":"Nebulex.Cache.stop/2","type":"callback"},{"doc":"Similar to all/2 but returns a lazy enumerable that emits all entries from the cache matching the given query. May raise Nebulex.QueryError if query validation fails. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :page_size - Positive integer (&gt;= 1) that defines the page size for the stream (defaults to 10). Examples # set some entries iex&gt; :ok = Enum.each(1..5, &amp;MyCache.set(&amp;1, &amp;1 * 2)) # stream all (with default params) iex&gt; MyCache.stream() |&gt; Enum.to_list() [1, 2, 3, 4, 5] # stream all entries and return values iex&gt; MyCache.stream(nil, return: :value, page_size: 3) |&gt; Enum.to_list() [2, 4, 6, 8, 10] # stream all entries and return objects iex&gt; stream = MyCache.stream(nil, return: :object, page_size: 3) iex&gt; [%Nebulex.Object{} | _] = Enum.to_list(stream) # additional built-in queries for Nebulex.Adapters.Local adapter all_unexpired_stream = MyCache.stream(:all_unexpired) all_expired_stream = MyCache.stream(:all_expired) # if we are using Nebulex.Adapters.Local adapter, the stored entry # is a tuple {key, value, version, expire_at}, then the match spec # could be something like: iex&gt; spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 5}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] iex&gt; MyCache.stream(spec, page_size: 3) |&gt; Enum.to_list() [{3, 6}, {4, 8}, {5, 10}] # the same previous query but using Ex2ms iex&gt; import Ex2ms Ex2ms iex&gt; spec = ...&gt; fun do ...&gt; {key, value, _, _} when value &gt; 5 -&gt; {key, value} ...&gt; end iex&gt; spec |&gt; MyCache.stream(page_size: 3) |&gt; Enum.to_list() [{3, 6}, {4, 8}, {5, 10}] To learn more, check out adapters&#39; documentation.","ref":"Nebulex.Cache.html#c:stream/2","title":"Nebulex.Cache.stream/2","type":"callback"},{"doc":"Returns and removes the object with key key in the cache. Returns nil if no result was found. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - Same as callback set/3. Examples iex&gt; MyCache.set(:a, 1) 1 iex&gt; MyCache.take(:a) 1 iex&gt; MyCache.take(:a) nil iex&gt; :a |&gt; MyCache.set(1, return: :key) |&gt; MyCache.take(return: :object) %Nebulex.Object{key: :a, value: 1, version: nil, expire_at: nil}","ref":"Nebulex.Cache.html#c:take/2","title":"Nebulex.Cache.take/2","type":"callback"},{"doc":"Similar to take/2 but raises KeyError if key is not found. Options See the &quot;Shared options&quot; section at the module documentation. Example MyCache.take!(:a)","ref":"Nebulex.Cache.html#c:take!/2","title":"Nebulex.Cache.take!/2","type":"callback"},{"doc":"Runs the given function inside a transaction. A successful transaction returns the value returned by the function. Options See the &quot;Shared options&quot; section at the module documentation. Examples MyCache.transaction fn -&gt; alice = MyCache.get(:alice) bob = MyCache.get(:bob) MyCache.set(:alice, %{alice | balance: alice.balance + 100}) MyCache.set(:bob, %{bob | balance: bob.balance + 100}) end # locking only the involved key (recommended): MyCache.transaction fn -&gt; alice = MyCache.get(:alice) bob = MyCache.get(:bob) MyCache.set(:alice, %{alice | balance: alice.balance + 100}) MyCache.set(:bob, %{bob | balance: bob.balance + 100}) end, keys: [:alice, :bob]","ref":"Nebulex.Cache.html#c:transaction/2","title":"Nebulex.Cache.transaction/2","type":"callback"},{"doc":"Updates the cached key with the given function. If key is present in Cache with value value, fun is invoked with argument value and its result is used as the new value of key. If key is not present in Cache, initial is inserted as the value of key. Options Besides the &quot;Shared options&quot; section at the module documentation, it accepts: :on_conflict - Same as callback set/3. Examples iex&gt; MyCache.update(:a, 1, &amp;(&amp;1 * 2)) 1 iex&gt; MyCache.update(:a, 1, &amp;(&amp;1 * 2)) 2 iex&gt; %Nebulex.Object{value: 4} = ...&gt; MyCache.update(:a, 1, &amp;(&amp;1 * 2), return: :object)","ref":"Nebulex.Cache.html#c:update/4","title":"Nebulex.Cache.update/4","type":"callback"},{"doc":"Updates (increment or decrement) the counter mapped to the given key. If incr &gt;= 0 then the current value is incremented by that amount, otherwise the current value is decremented. If incr is not a valid integer, then an ArgumentError exception is raised. Options See the &quot;Shared options&quot; section at the module documentation. Note that for this function, option :version is ignored. Examples iex&gt; MyCache.update_counter(:a) 1 iex&gt; MyCache.update_counter(:a, 2) 3 iex&gt; MyCache.update_counter(:a, -1) 2 iex&gt; %Nebulex.Object{key: :a, value: 2} = ...&gt; MyCache.update_counter(:a, 0, return: :object)","ref":"Nebulex.Cache.html#c:update_counter/3","title":"Nebulex.Cache.update_counter/3","type":"callback"},{"doc":"Cache entries","ref":"Nebulex.Cache.html#t:entries/0","title":"Nebulex.Cache.entries/0","type":"type"},{"doc":"Object key","ref":"Nebulex.Cache.html#t:key/0","title":"Nebulex.Cache.key/0","type":"type"},{"doc":"Cache object","ref":"Nebulex.Cache.html#t:object/0","title":"Nebulex.Cache.object/0","type":"type"},{"doc":"Cache action options","ref":"Nebulex.Cache.html#t:opts/0","title":"Nebulex.Cache.opts/0","type":"type"},{"doc":"Return alternatives (value is the default)","ref":"Nebulex.Cache.html#t:return/0","title":"Nebulex.Cache.return/0","type":"type"},{"doc":"","ref":"Nebulex.Cache.html#t:t/0","title":"Nebulex.Cache.t/0","type":"type"},{"doc":"Object value","ref":"Nebulex.Cache.html#t:value/0","title":"Nebulex.Cache.value/0","type":"type"},{"doc":"Function decorators which provide a way of annotating functions to be cached or evicted. By means of these decorators, it is possible the implementation of cache usage patterns like Read-through, Write-through, Cache-as-SoR, etc. Shared Options All of the caching macros below accept the following options: :cache - Defines what cache to use (required). Raises ArgumentError if the option is not present. :key - Defines the cache access key (optional). If this option is not present, a default key is generated by hashing a two-elements tuple; first element is the function&#39;s name and the second one the list of arguments (e.g: :erlang.phash2({name, args})). :opts - Defines the cache options that will be passed as argument to the invoked cache function (optional). :match - Defines a function that takes one argument and will be used to decide if the cache should be updated or not (optional). If this option is not present, the value will always be updated. Does not have any effect upon eviction since values are always evicted before executing the function logic. Example Suppose we are using Ecto and we want to define some caching functions in the context MyApp.Accounts. defmodule MyApp.Accounts do use Nebulex.Caching.Decorators import Ecto.Query alias MyApp.Accounts.User alias MyApp.Cache alias MyApp.Repo @decorate cache(cache: Cache, key: {User, id}, opts: [ttl: 3600]) def get_user!(id) do Repo.get!(User, id) end @decorate cache(cache: Cache, key: {User, clauses}) def get_user_by!(clauses) do Repo.get_by!(User, clauses) end @decorate cache(cache: Cache) def users_by_segment(segment \\\\\\\\ &quot;standard&quot;) do query = from(q in User, where: q.segment == ^segment) Repo.all(query) end @decorate cache(cache: Cache, key: {User, :latest}, match: &amp;(not is_nil(&amp;1))) def get_newest_user() do Repo.get_newest(User) end @decorate update(cache: Cache, key: {User, user.id}) def update_user!(%User{} = user, attrs) do user |&gt; User.changeset(attrs) |&gt; Repo.update!() end @decorate evict(cache: Cache, keys: [{User, user.id}, {User, [username: user.username]}]) def delete_user(%User{} = user) do Repo.delete(user) end end","ref":"Nebulex.Caching.Decorators.html","title":"Nebulex.Caching.Decorators","type":"module"},{"doc":"","ref":"Nebulex.Caching.Decorators.html#cache/1","title":"Nebulex.Caching.Decorators.cache/1","type":"macro"},{"doc":"Provides a way of annotating functions to be cached (cacheable aspect). The returned value by the code block is cached if it doesn&#39;t exist already in cache, otherwise, it is returned directly from cache and the code block is not executed. Options See the &quot;Shared options&quot; section at the module documentation. Examples defmodule MyApp.Example do use Nebulex.Caching.Decorators alias MyApp.Cache @decorate cache(cache: Cache, key: name) def get_by_name(name, age) do # your logic (maybe the loader to retrieve the value from the SoR) end @decorate cache(cache: Cache, key: age, opts: [ttl: 3600]) def get_by_age(age) do # your logic (maybe the loader to retrieve the value from the SoR) end @decorate cache(cache: Cache) def all(query) do # your logic (maybe the loader to retrieve the value from the SoR) end @decorate cache(cache: Cache, key: {User, :latest}, match: &amp;(not is_nil(&amp;1))) def get_newest_user() do Repo.get_newest(User) end end The Read-through pattern is supported by this decorator. The loader to retrieve the value from the system-of-record (SoR) is your function&#39;s logic and the rest is provided by the macro under-the-hood.","ref":"Nebulex.Caching.Decorators.html#cache/3","title":"Nebulex.Caching.Decorators.cache/3","type":"function"},{"doc":"","ref":"Nebulex.Caching.Decorators.html#evict/1","title":"Nebulex.Caching.Decorators.evict/1","type":"macro"},{"doc":"Provides a way of annotating functions to be evicted (eviction aspect). On function&#39;s completion, the given key or keys (depends on the :key and :keys options) are deleted from the cache. Options :keys - Defines the set of keys meant to be evicted from cache on function completion. :all_entries - Defines if all entries must be removed on function completion. Defaults to false. See the &quot;Shared options&quot; section at the module documentation. Examples defmodule MyApp.Example do use Nebulex.Caching.Decorators alias MyApp.Cache @decorate evict(cache: Cache, key: name) def evict(name) do # your logic (maybe write/delete data to the SoR) end @decorate evict(cache: Cache, keys: [name, id]) def evict_many(name) do # your logic (maybe write/delete data to the SoR) end @decorate evict(cache: Cache, all_entries: true) def evict_all(name) do # your logic (maybe write/delete data to the SoR) end end The Write-through pattern is supported by this decorator. Your function provides the logic to write data to the system-of-record (SoR) and the rest is provided by the decorator under-the-hood. But in contrast with update decorator, when the data is written to the SoR, the key for that value is deleted from cache instead of updated.","ref":"Nebulex.Caching.Decorators.html#evict/3","title":"Nebulex.Caching.Decorators.evict/3","type":"function"},{"doc":"","ref":"Nebulex.Caching.Decorators.html#update/1","title":"Nebulex.Caching.Decorators.update/1","type":"macro"},{"doc":"Provides a way of annotating functions to be evicted; but updating the cached key instead of deleting it. The content of the cache is updated without interfering with the function execution. That is, the method would always be executed and the result cached. The difference between cache/3 and update/3 is that cache/3 will skip running the function if the key exists in the cache, whereas update/3 will actually run the function and then put the result in the cache. Options See the &quot;Shared options&quot; section at the module documentation. Examples defmodule MyApp.Example do use Nebulex.Caching.Decorators alias MyApp.Cache @decorate update(cache: Cache, key: name) def update(name) do # your logic (maybe write data to the SoR) end @decorate update(cache: Cache, opts: [ttl: 3600]) def update_with_ttl(name) do # your logic (maybe write data to the SoR) end @decorate update(cache: Cache, match: &amp;match_function/1) def update_when_not_nil() do # your logic (maybe write data to the SoR) end defp match_function(value) do # your condition to skip updating the cache edn end The Write-through pattern is supported by this decorator. Your function provides the logic to write data to the system-of-record (SoR) and the rest is provided by the decorator under-the-hood.","ref":"Nebulex.Caching.Decorators.html#update/3","title":"Nebulex.Caching.Decorators.update/3","type":"function"},{"doc":"This module specifies the behaviour for pre/post hooks callbacks. These functions are defined in order to intercept any cache operation and be able to execute a set of actions before and/or after the operation takes place. Execution modes It is possible to setup the mode how the hooks are evaluated. The pre_hooks/0 and post_hooks/0 callbacks must return a tuple {mode, hook_funs}, where the first element mode is the one that defines the execution mode. The available modes are: :async - (the default) all hooks are evaluated asynchronously (in parallel) and their results are ignored. :sync - hooks are evaluated synchronously (sequentially) and their results are ignored. :pipe - similar to :sync but each hook result is passed to the next one and so on, until the last hook evaluation is returned. Example defmodule MyApp.MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local def pre_hooks do {:async, [... your pre hook functions ...]} end def post_hooks do {:pipe, [... your post hook functions ...]} end end","ref":"Nebulex.Hook.html","title":"Nebulex.Hook","type":"behaviour"},{"doc":"Evaluates the hooks according to the given execution mode.","ref":"Nebulex.Hook.html#eval/3","title":"Nebulex.Hook.eval/3","type":"function"},{"doc":"Returns a list of hook functions that will be executed after invoke the cache action. Examples defmodule MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local def post_hooks do {:pipe, [&amp;post_hook/2]} end def post_hook(result, {_, :set, _} = call) do send(:hooked_cache, call) end def post_hook(_, _) do :noop end end","ref":"Nebulex.Hook.html#c:post_hooks/0","title":"Nebulex.Hook.post_hooks/0","type":"callback"},{"doc":"Returns a list of hook functions that will be executed before invoke the cache action. Examples defmodule MyCache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local def pre_hooks do pre_hook = fn (result, {_, :get, _} = call) -&gt; # do your stuff ... (result, _) -&gt; result end {:async, [pre_hook]} end end","ref":"Nebulex.Hook.html#c:pre_hooks/0","title":"Nebulex.Hook.pre_hooks/0","type":"callback"},{"doc":"Defines a cache command","ref":"Nebulex.Hook.html#t:command/0","title":"Nebulex.Hook.command/0","type":"type"},{"doc":"Defines the hook callback function","ref":"Nebulex.Hook.html#t:hook_fun/0","title":"Nebulex.Hook.hook_fun/0","type":"type"},{"doc":"Hook execution mode","ref":"Nebulex.Hook.html#t:mode/0","title":"Nebulex.Hook.mode/0","type":"type"},{"doc":"Defines a Cache Object. This is the struct used by the caches to store and retrieve data.","ref":"Nebulex.Object.html","title":"Nebulex.Object","type":"module"},{"doc":"","ref":"Nebulex.Object.html#decode/2","title":"Nebulex.Object.decode/2","type":"function"},{"doc":"","ref":"Nebulex.Object.html#encode/2","title":"Nebulex.Object.encode/2","type":"function"},{"doc":"Returns the UNIX timestamp in seconds for the given ttl. Example iex&gt; expire_at = Nebulex.Object.expire_at(10) iex&gt; expire_at - Nebulex.Object.ts() 10","ref":"Nebulex.Object.html#expire_at/1","title":"Nebulex.Object.expire_at/1","type":"function"},{"doc":"Returns whether the given object has expired or not. Example iex&gt; Nebulex.Object.expired?(%Nebulex.Object{}) false","ref":"Nebulex.Object.html#expired?/1","title":"Nebulex.Object.expired?/1","type":"function"},{"doc":"Returns the remaining time to live for the given timestamp. Example iex&gt; expire_at = Nebulex.Object.expire_at(10) iex&gt; Nebulex.Object.remaining_ttl(expire_at) 10","ref":"Nebulex.Object.html#remaining_ttl/1","title":"Nebulex.Object.remaining_ttl/1","type":"function"},{"doc":"Wrapper for DateTime.to_unix/2. Example iex&gt; 1_464_096_368 |&gt; DateTime.from_unix!() |&gt; Nebulex.Object.ts() 1464096368","ref":"Nebulex.Object.html#ts/1","title":"Nebulex.Object.ts/1","type":"function"},{"doc":"","ref":"Nebulex.Object.html#t:t/0","title":"Nebulex.Object.t/0","type":"type"},{"doc":"Version handler for cached objects. The purpose of this module is to allow users implement custom version generator and validator. This interface is used by the adapters to generate and/or validate object versions. Example defmodule MyApp.Versioner do @behaviour Nebulex.Object.Version def generate(nil), do: 0 def generate(%Object{version: version}) do version + 1 end end If you are going to rely on this feature, it is recommended to use a good version generation algorithm (e.g.: Vector Clocks).","ref":"Nebulex.Object.Version.html","title":"Nebulex.Object.Version","type":"behaviour"},{"doc":"Generates a new version for the current cached object.","ref":"Nebulex.Object.Version.html#c:generate/1","title":"Nebulex.Object.Version.generate/1","type":"callback"},{"doc":"Validates version conflicts for the given cached object_or_key. This function is used by cache&#39;s adapters which implement optimistic locking via object&#39;s version. For more information, check out adapters&#39; implementation.","ref":"Nebulex.Object.Version.html#validate/3","title":"Nebulex.Object.Version.validate/3","type":"function"},{"doc":"RPC utilities for distributed task execution. This module uses supervised tasks underneath Task.Supervisor.","ref":"Nebulex.RPC.html","title":"Nebulex.RPC","type":"module"},{"doc":"Evaluates apply(mod, fun, args) on node node and returns the corresponding evaluation result, or {:badrpc, reason} if the call fails. A timeout, in milliseconds or :infinity, can be given with a default value of 5000. It uses Task.await/2 internally. Example iex&gt; Nebulex.RPC.call(:my_task_sup, :node1, Kernel, :to_string, [1]) &quot;1&quot;","ref":"Nebulex.RPC.html#call/6","title":"Nebulex.RPC.call/6","type":"function"},{"doc":"In contrast to a regular single-node RPC, a multicall is an RPC that is sent concurrently from one client to multiple servers. The function evaluates apply(mod, fun, args) on each node_group entry and collects the answers. Then, evaluates the reducer function (set in the opts) on each answer. This function is similar to :rpc.multicall/5. Options timeout - A timeout, in milliseconds or :infinity, can be given with a default value of 5000. It uses Task.yield_many/2 internally. reducer - Reducer function to be executed on each collected result. (check out reducer type). Example iex&gt; Nebulex.RPC.multi_call( ...&gt; :my_task_sup, ...&gt; %{ ...&gt; node1: {Kernel, :to_string, [1]}, ...&gt; node2: {Kernel, :to_string, [2]} ...&gt; }, ...&gt; timeout: 10_000, ...&gt; reducer: { ...&gt; [], ...&gt; fn ...&gt; {:ok, res}, _node_callback, acc -&gt; ...&gt; [res | acc] ...&gt; ...&gt; {:error, _}, _node_callback, acc -&gt; ...&gt; acc ...&gt; end ...&gt; } ...&gt; ) [&quot;1&quot;, &quot;2&quot;]","ref":"Nebulex.RPC.html#multi_call/3","title":"Nebulex.RPC.multi_call/3","type":"function"},{"doc":"Similar to multi_call/3 but the same node_callback (given by module, fun, args) is executed on all nodes; Internally it creates a node_group with the same node_callback for each node. Options Same options as multi_call/3. Example iex&gt; Nebulex.RPC.multi_call( ...&gt; :my_task_sup, ...&gt; [:node1, :node2], ...&gt; Kernel, ...&gt; :to_string, ...&gt; [1], ...&gt; timeout: 5000, ...&gt; reducer: { ...&gt; [], ...&gt; fn ...&gt; {:ok, res}, _node_callback, acc -&gt; ...&gt; [res | acc] ...&gt; ...&gt; {:error, _}, _node_callback, acc -&gt; ...&gt; acc ...&gt; end ...&gt; } ...&gt; ) [&quot;1&quot;, &quot;1&quot;]","ref":"Nebulex.RPC.html#multi_call/6","title":"Nebulex.RPC.multi_call/6","type":"function"},{"doc":"Task callback","ref":"Nebulex.RPC.html#t:callback/0","title":"Nebulex.RPC.callback/0","type":"type"},{"doc":"Group entry: node -&gt; callback","ref":"Nebulex.RPC.html#t:node_callback/0","title":"Nebulex.RPC.node_callback/0","type":"type"},{"doc":"Node group","ref":"Nebulex.RPC.html#t:node_group/0","title":"Nebulex.RPC.node_group/0","type":"type"},{"doc":"Reducer spec","ref":"Nebulex.RPC.html#t:reducer/0","title":"Nebulex.RPC.reducer/0","type":"type"},{"doc":"Raised at runtime when a key already exists in cache.","ref":"Nebulex.KeyAlreadyExistsError.html","title":"Nebulex.KeyAlreadyExistsError","type":"exception"},{"doc":"Raised at runtime when the query is invalid.","ref":"Nebulex.QueryError.html","title":"Nebulex.QueryError","type":"exception"},{"doc":"Raised at runtime when a RPC multi_call error occurs.","ref":"Nebulex.RPCMultiCallError.html","title":"Nebulex.RPCMultiCallError","type":"exception"},{"doc":"Raised at runtime when there is a version conflict between the required object version and the cached object version.","ref":"Nebulex.VersionConflictError.html","title":"Nebulex.VersionConflictError","type":"exception"},{"doc":"Getting Started This guide is an introduction to Nebulex, a local and distributed caching library for Elixir. Nebulex API is pretty much inspired by Ecto, taking advantage of its simplicity, flexibility and pluggable architecture. In the same way as Ecto, developers can provide their own cache (adapter) implementations. In this guide, we&#39;re going to learn some basics about Nebulex, such as set, retrieve and destroy cache entries (key/value pairs).","ref":"getting-started.html","title":"Getting Started","type":"extras"},{"doc":"Let&#39;s start creating a new Elixir application by running this command: mix new blog --sup The --sup option ensures that this application has a supervision tree, which will be needed by Nebulex later on. To add Nebulex to this application, there are a few steps that we need to take. The first step will be adding Nebulex to our mix.exs file, which we&#39;ll do by changing the deps definition in that file to this: defp deps do [ {:nebulex, &quot;~&gt; 1.2.0&quot;} ] end To install these dependencies, we will run this command: mix deps.get For the second step we need to setup some configuration for Nebulex so that we can perform actions on a cache from within the application&#39;s code. We can set up this configuration by running this command: mix nebulex.gen.cache -c Blog.Cache This command will generate the configuration required to use our cache. The first bit of configuration is in config/config.exs: config :blog, Blog.Cache, gc_interval: 86_400 # 24 hrs The Blog.Cache module is defined in lib/blog/cache.ex by our mix nebulex.gen.cache command: defmodule Blog.Cache do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Local end This module is what we&#39;ll be using to interact with the cache. It uses the Nebulex.Cache module, and the otp_app tells Nebulex which Elixir application it can look for cache configuration in. In this case, we&#39;ve specified that it is the :blog application where Nebulex can find that configuration and so Nebulex will use the configuration that was set up in config/config.exs. The final piece of configuration is to setup the Blog.Cache as a supervisor within the application&#39;s supervision tree, which we can do in lib/blog/application.ex (or lib/blog.ex for elixir versions &lt; 1.4.0), inside the start/2 function: Elixir &lt; 1.5.0: def start(_type, _args) do import Supervisor.Spec, warn: false children = [ supervisor(Blog.Cache, []), ] ... Elixir &gt;= 1.5.0: def start(_type, _args) do children = [ Blog.Cache ] ... This piece of configuration will start the Nebulex process which receives and executes our application&#39;s commands. Without it, we wouldn&#39;t be able to use the cache at all! We&#39;ve now configured our application so that it&#39;s able to execute commands against our cache.","ref":"getting-started.html#adding-nebulex-to-an-application","title":"Getting Started - Adding Nebulex to an application","type":"extras"},{"doc":"We can insert a new entries into our blog cache with this code: user = %{id: 1, username: &quot;cabol&quot;, email: &quot;cabol@email.com&quot;} Blog.Cache.set(user[:id], user) To insert the data into our cache, we call set on Blog.Cache. This function tells Nebulex that we want to insert a new key/value entry into the cache corresponding Blog.Cache. A successful set will return (by default) the inserted value, like so: %{id: 1, username: &quot;cabol&quot;, email: &quot;cabol@email.com&quot;} But, using the option :return we can ask for return either the value, key or the entire Nebulex.Object, like so: Blog.Cache.set(&quot;foo&quot;, &quot;bar&quot;, return: :key) Blog.Cache.set(&quot;foo&quot;, &quot;bar&quot;, return: :object) Blog.Cache.set(&quot;foo&quot;, &quot;bar&quot;, return: :value) # Default Add and Replace As we saw previously, set creates a new entry in cache if it doesn&#39;t exist, or overrides it if it does exist (including the :ttl). However, there might be circumstances where we want to set the entry only if it doesn&#39;t exit or the other way around, this is where add, add!, replace and replace! functions come in. Let&#39;s try add and add! functions: {:ok, &quot;value&quot;} = Blog.Cache.add(&quot;new&quot;, &quot;value&quot;) &quot;value&quot; = Blog.Cache.add!(&quot;new2&quot;, &quot;value&quot;) # returns `:error` because the `key` already exists :error = Blog.Cache.add(&quot;new&quot;, &quot;value&quot;) # same as previous one but raises [`Nebulex.KeyAlreadyExistsError`](Nebulex.KeyAlreadyExistsError.html) Blog.Cache.add!(&quot;new&quot;, &quot;new value&quot;) Now replace and replace! functions: {:ok, &quot;new value&quot;} = Blog.Cache.replace(&quot;new&quot;, &quot;new value&quot;) &quot;new value&quot; = Blog.Cache.replace!(&quot;new2&quot;, &quot;new value&quot;) # returns `:error` because the `key` doesn&#39;t exist :error = Blog.Cache.replace(&quot;another&quot;, &quot;new value&quot;) # same as previous one but raises [`KeyError`](https://hexdocs.pm/elixir/KeyError.html) Blog.Cache.replace!(&quot;another&quot;, &quot;new value&quot;) # update only the TTL without alter the current value (value is set to nil) Blog.Cache.replace!(&quot;existing key&quot;, nil, ttl: 60) # updating both, value and TTL Blog.Cache.replace!(&quot;existing key&quot;, &quot;value&quot;, ttl: 60)","ref":"getting-started.html#inserting-entries","title":"Getting Started - Inserting entries","type":"extras"},{"doc":"First, let&#39;s create some data as we learned before. users = [ %{id: 1, first_name: &quot;Galileo&quot;, last_name: &quot;Galilei&quot;}, %{id: 2, first_name: &quot;Charles&quot;, last_name: &quot;Darwin&quot;}, %{id: 3, first_name: &quot;Albert&quot;, last_name: &quot;Einstein&quot;} ] Enum.each(users, fn(user) -&gt; Blog.Cache.set(user[:id], user) end) This code will create three new users in our cache. Fetching a single entry Let’s start off with fetching data by the key, which is the most basic and common operation to retrieve data from a cache. Blog.Cache.get(1) for key &lt;- 1..3, do: Blog.Cache.get(key) By default, get returns the value associated to the given key, but in the same way as set, we can ask for return either the key, value or object. for key &lt;- 1..3, do: Blog.Cache.get(key, return: :key) for key &lt;- 1..3, do: Blog.Cache.get(key, return: :object) Additionally, there is a function has_key? to check if a key exist in cache: Blog.Cache.has_key?(1) It returns true if the ket exist and false otherwise. Fetch and/or stream multiple entires Nebulex provides functions to fetch or stream all entries from cache matching the given query. To fetch all entries from cache: # by default, returns all keys Blog.Cache.all() # fetch all entries and return values Blog.Cache.all(nil, return: :value) # fetch all entries and return objects Blog.Cache.all(nil, return: :object) # built-in queries in [`Nebulex.Adapters.Local`](Nebulex.Adapters.Local.html) adapter Blog.Cache.all(nil) Blog.Cache.all(:all_unexpired) Blog.Cache.all(:all_expired) # if we are using [`Nebulex.Adapters.Local`](Nebulex.Adapters.Local.html) adapter, the stored entry # is a tuple `{key, value, version, expire_at}`, then the match spec # could be something like: spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 10}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] Blog.Cache.all(spec) # using Ex2ms import Ex2ms spec = fun do {key, value, _, _} when value &gt; 10 -&gt; {key, value} end Blog.Cache.all(spec) In the same way, we can stream all entries: Blog.Cache.stream() Blog.Cache.stream(nil, page_size: 3, return: :value) Blog.Cache.stream(nil, page_size: 3, return: :object) # using [`Nebulex.Adapters.Local`](Nebulex.Adapters.Local.html) adapter spec = [{{:&quot;$1&quot;, :&quot;$2&quot;, :_, :_}, [{:&gt;, :&quot;$2&quot;, 10}], [{{:&quot;$1&quot;, :&quot;$2&quot;}}]}] Blog.Cache.stream(spec, page_size: 3) # using Ex2ms import Ex2ms spec = fun do {key, value, _, _} when value &gt; 10 -&gt; {key, value} end Blog.Cache.stream(spec, page_size: 3)","ref":"getting-started.html#retrieving-entries","title":"Getting Started - Retrieving entries","type":"extras"},{"doc":"If you want to generate the new entry based on the current one, you can use get and then set, if you don&#39;t care about the current cached value, you can use only set (or replace), like so: v1 = Blog.Cache.get(1) Blog.Cache.set(1, %{v1 | first_name: &quot;Nebulex&quot;}) # In case you don&#39;t care about an existing entry, you can just override the # existing one (if it exists); remember, `set` is an idempotent operation Blog.Cache.set(1, &quot;anything&quot;) However, Nebulex provides update and get_and_update functions to update an entry value based on current one, for example: initial = %{id: 1, first_name: &quot;&quot;, last_name: &quot;&quot;} # using `get_and_update` Blog.Cache.get_and_update(1, fn v -&gt; if v, do: {v, %{v | first_name: &quot;X&quot;}}, else: {v, initial} end) # using `update` Blog.Cache.update(1, initial, &amp;(%{&amp;1 | first_name: &quot;Y&quot;}))","ref":"getting-started.html#updating-entries","title":"Getting Started - Updating entries","type":"extras"},{"doc":"Nebulex also provides the function update_counter in order to handle counters, increments and decrements; by default, a counter is set/initialized to 0. Let&#39;s see how counters works: # by default, the counter is incremented by 1 Blog.Cache.update_counter(:my_counter) # but we can also provide a custom increment value Blog.Cache.update_counter(:my_counter, 5) # to decrement the counter, just pass a negative value Blog.Cache.update_counter(:my_counter, -5)","ref":"getting-started.html#updating-counters","title":"Getting Started - Updating counters","type":"extras"},{"doc":"We’ve now covered inserting (set), reading (get, get!, all) and updating objects. Now let&#39;s see how to delete an object using Nebulex; the basic and/or main way to do so is through delete/2 function. Blog.Cache.delete(1) By default, delete returns the key either if success or not, however, it is possible also to pass the :return option. Take There is another way to delete an entry and at the same time to retrieve it, the function to achieve this is take or take!, this is an example: Blog.Cache.take(1) Blog.Cache.take(2, return: :key) Blog.Cache.take!(3, return: :object) # returns `nil` if `key` doesn&#39;t exist nil = Blog.Cache.take(&quot;nonexistent&quot;) # same as previous one but raises [`KeyError`](https://hexdocs.pm/elixir/KeyError.html) Blog.Cache.take!(&quot;nonexistent&quot;) Similar to set and get, take returns the value by default, but you can request to return either the key, value or object. Flush Nebulex also provides a function to flush all cache entries, like so: Blog.Cache.flush()","ref":"getting-started.html#deleting-entries","title":"Getting Started - Deleting entries","type":"extras"},{"doc":"The last thing we’ll cover in this guide is how to retrieve information about cached objects or the cache itself. Object Info To retrieve the TTL of an object: Blog.Cache.object_info(&quot;mykey&quot;, :ttl) And if we want to retrieve the version: Blog.Cache.object_info(&quot;mykey&quot;, :version) Cache Info To get the total number of cached objects: Blog.Cache.size()","ref":"getting-started.html#info","title":"Getting Started - Info","type":"extras"},{"doc":"Nebulex provides the adapter Nebulex.Adapters.Partitioned, which allows to setup a partitioned cache topology. Let&#39;s setup our partitioned cache by running this command: mix nebulex.gen.cache -c Blog.PartitionedCache -a Nebulex.Adapters.Partitioned This command will generate the configuration required to use our partitioned cache; it is defined in config/config.exs: config :blog, Blog.PartitionedCache, primary: :YOUR_LOCAL_CACHE, node_selector: Nebulex.Adapters.Partitioned Replace the local config value :YOUR_LOCAL_CACHE by an existing local cache (e.g.: we can set the local cache we created previously), or create a new one (we will create a new one within Blog.PartitionedCache). The Blog.PartitionedCache module is defined in lib/blog/partitioned_cache.ex by our mix nebulex.gen.cache command: defmodule Blog.PartitionedCache do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Partitioned end As mentioned previously, let&#39;s add the local backend (local cache): defmodule Blog.PartitionedCache do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Partitioned defmodule Primary do use Nebulex.Cache, otp_app: :blog, adapter: Nebulex.Adapters.Local end end Now we have to add the new local cache to the config, and also replace the local config value :YOUR_LOCAL_CACHE by our new local cache in the partitioned cache config. # Local backend for the partitioned cache config :blog, Blog.PartitionedCache.Primary, gc_interval: 86_400 # 24 hrs # Partitioned Cache config :blog, Blog.PartitionedCache, primary: Blog.PartitionedCache.Primary, node_selector: Nebulex.Adapters.Partitioned And remember to setup the Blog.PartitionedCache and its local backend Blog.PartitionedCache.Primary as supervisors within the application&#39;s supervision tree (such as we did it previously): Elixir &lt; 1.5.0: def start(_type, _args) do import Supervisor.Spec, warn: false children = [ supervisor(Blog.Cache, []), # Previous created local cache supervisor(Blog.PartitionedCache, []), # Partitioned cache supervisor(Blog.PartitionedCache.Primary, []) # Local cache that will be used by the partitioned cache ] ... Elixir &gt;= 1.5.0: def start(_type, _args) do import Supervisor.Spec, warn: false children = [ Blog.Cache, # Previous created local cache Blog.PartitionedCache, # Partitioned cache Blog.PartitionedCache.Primary # Local cache that will be used by the partitioned cache ] ... Now we are ready to start using our partitioned cache! Timeout option The Nebulex.Adapters.Partitioned supports :timeout option, it is a value in milliseconds for the command that will be executed. Blog.PartitionedCache.get(&quot;foo&quot;, timeout: 10) # if the timeout is exceeded, then the current process will exit Blog.PartitionedCache.set(&quot;foo&quot;, &quot;bar&quot;, timeout: 10) # ** (EXIT) time out To learn more about how partitioned cache works, please check Nebulex.Adapters.Partitioned documentation, and also it is recommended see the partitioned cache example","ref":"getting-started.html#partitioned-cache","title":"Getting Started - Partitioned Cache","type":"extras"},{"doc":"Nebulex also provides the adapter Nebulex.Adapters.Multilevel, which allows to setup a multi-level caching hierarchy. First, let&#39;s create a multi-level cache module: mix nebulex.gen.cache -c Blog.Multilevel -a Nebulex.Adapters.Multilevel This command will generate the configuration required to use our multilevel cache; it is defined in config/config.exs: config :blog, Blog.Multilevel, cache_model: :inclusive, levels: [] Next step is to set the levels config value, with the caches that will be part of the caching hierarchy. Let&#39;s suppose we want a two levels cache, L1 and L2 cache, where L1 (first level) is our local cache and L2 (second level) the partitioned cache. Therefore, the configuration would be like so: config :blog, Blog.Multilevel, cache_model: :inclusive, levels: [Blog.Cache, Blog.PartitionedCache] Note that the Blog.PartitionedCache.Primary cannot be part of the levels, since it is the backend used by Blog.PartitionedCache behind scenes. And remember to setup the Blog.Multilevel as a supervisor within the application&#39;s supervision tree (such as we did it previously): Elixir &lt; 1.5.0: def start(_type, _args) do import Supervisor.Spec, warn: false children = [ supervisor(Blog.Cache, []), # Previous created local cache supervisor(Blog.PartitionedCache, []), # Partitioned cache supervisor(Blog.PartitionedCache.Primary, []), # Local cache that will be used by the partitioned cache supervisor(Blog.MultilevelCache, []) # Multilevel cache ] ... Elixir &gt;= 1.5.0: def start(_type, _args) do import Supervisor.Spec, warn: false children = [ Blog.Cache, # Previous created local cache Blog.PartitionedCache, # Partitioned cache Blog.PartitionedCache.Primary, # Local cache that will be used by the partitioned cache Blog.Multilevel # Multilevel cache ] ... Let&#39;s try it out! Insert some date into the partitioned cache: iex&gt; Blog.PartitionedCache.set(&quot;foo&quot;, &quot;bar&quot;) &quot;bar&quot; iex&gt; Blog.Cache.get(&quot;foo&quot;) nil iex&gt; Blog.PartitionedCache.get(&quot;foo&quot;) &quot;bar&quot; Now let&#39;s retrieve the data but using the multi-level cache: iex&gt; Blog.Multilevel.get(&quot;foo&quot;) &quot;bar&quot; iex&gt; Blog.Cache.get(&quot;foo&quot;) &quot;bar&quot; iex&gt; Blog.PartitionedCache.get(&quot;foo&quot;) &quot;bar&quot; As you can see the date is now cached in out local cache Blog.Cache, the multi-level cache did the work. To learn more about how multilevel-cache works, please check Nebulex.Adapters.Multilevel documentation, and also it is recommended see the near cache example","ref":"getting-started.html#multilevel-cache","title":"Getting Started - Multilevel Cache","type":"extras"},{"doc":"Nebulex.Caching DSL - Tailored DSL to implement different cache usage patterns. Pre and Post Hooks - Ability to hook any function call for a cache and add custom logic before and/or after function execution.","ref":"getting-started.html#other-important-guides","title":"Getting Started - Other important guides","type":"extras"},{"doc":"Pre/Post Hooks Pre/Post hooks are functions that you define and that you direct to execute before or after particular methods. Hooks mechanism is very powerful but dangerous at the same time, so you have to be careful. When we define a cache, we are able to override pre_hooks/0 and post_hooks/0 functions by providing ours; these callbacks are defined by c:Nebulex.Cache.Hook behaviour. Let&#39;s check out the example just below in order to understand better how it works.","ref":"hooks.html","title":"Pre/Post Hooks","type":"extras"},{"doc":"Suppose we want to trace all cache calls (before and after they are called) by logging them. In this case, we need to provide a pre and post hook to log these calls. Supposing you have an app already and it is also setup, your cache might looks like this: defmodule MyApp.Cache do use Nebulex.Cache, otp_app: :my_app, adapter: Nebulex.Adapters.Local require Logger def pre_hooks do log = fn result, call -&gt; Logger.debug &quot;PRE: \#{inspect call} ==&gt; \#{inspect result}&quot; end {:async, [log]} end def post_hooks do log = fn result, call -&gt; Logger.debug &quot;POST: \#{inspect call} ==&gt; \#{inspect result}&quot; end {:pipe, [log]} end end See how the pre_hooks/0 and post_hooks/0 are overridden providing our own pre/post hooks for tracing/logging purposes. Then, if we open a console and try out: iex(1)&gt; MyApp.Cache.set 1, 1 10:19:47.736 [debug] PRE: {MyApp.Cache, :set, [1, 1, []]} 1 10:19:47.736 [debug] POST: {MyApp.Cache, :set, [1, 1, []]} ==&gt; 1 iex(2)&gt; MyApp.Cache.get 1 10:20:14.941 [debug] PRE: {MyApp.Cache, :get, [1, []]} 1 10:20:14.941 [debug] POST: {MyApp.Cache, :get, [1, []]} ==&gt; 1 See how our hooks are logging all cache calls!","ref":"hooks.html#logging-hooks-example","title":"Pre/Post Hooks - Logging Hooks Example","type":"extras"},{"doc":"It is possible to setup the mode how the hooks are evaluated. The available modes are: :async - all hooks are evaluated asynchronously (in parallel) and their results are ignored. :sync - hooks are evaluated synchronously (sequentially) and their results are ignored. :pipe - similar to :sync but each hook result is passed to the next one and so on, until the last hook evaluation is returned. You have to be careful with :pipe option, because the result of the hook is passed to the next one and so on, and the final result will be the result of the last hook. So, as you can see, the result can be manipulated, but this behaviour only affects the post hooks, because the result of pre hooks is ignored – the result is piped between pre hooks, but the final result never arrives to the cache action.","ref":"hooks.html#configuration","title":"Pre/Post Hooks - Configuration","type":"extras"},{"doc":"Cache Usage Patterns via Nebulex.Caching.Decorators There are several common access patterns when using a cache. Nebulex supports most of these patterns by means of Nebulex.Caching.Decorators. Most of the following documentation about caching patterns it has been taken from EHCache Doc","ref":"caching-decorators.html","title":"Cache Usage Patterns via Nebulex.Caching.Decorators","type":"extras"},{"doc":"With the cache-aside pattern, application code uses the cache directly. This means that application code which accesses the system-of-record (SoR) should consult the cache first, and if the cache contains the data, then return the data directly from the cache, bypassing the SoR. Otherwise, the application code must fetch the data from the system-of-record, store the data in the cache, and then return it. When data is written, the cache must be updated along with the system-of-record. Reading values if value = MyCache.get(key) do value else value = SoR.get(key) # maybe Ecto? MyCache.set(key, value) end Writing values MyCache.set(key, value) SoR.insert(key, value) # maybe Ecto? As you may have noticed, this is the default behavior for most of the caches, we have to interact directly with the cache as well as the SoR (most likely the Database).","ref":"caching-decorators.html#cache-aside","title":"Cache Usage Patterns via Nebulex.Caching.Decorators - Cache-aside","type":"extras"},{"doc":"The cache-as-SoR pattern implies using the cache as though it were the primary system-of-record (SoR). The pattern delegates SoR reading and writing activities to the cache, so that application code is (at least directly) absolved of this responsibility. To implement the cache-as-SoR pattern, use a combination of the following read and write patterns: Read-through Write-through Advantages of using the cache-as-SoR pattern are: Less cluttered application code (improved maintainability through centralized SoR read/write operations) Choice of write-through or write-behind strategies on a per-cache basis Allows the cache to solve the thundering-herd problem A disadvantage of using the cache-as-SoR pattern is: Less directly visible code-path But how to get all this out-of-box? This is where the Decorators comes in. It provides a set of macros to abstract most of the logic behind Read-through and Write-through patterns and make the implementation extremely easy. But let&#39;s go over these patterns more in detail and how to implement them using Nebulex Decorators.","ref":"caching-decorators.html#cache-as-sor","title":"Cache Usage Patterns via Nebulex.Caching.Decorators - Cache-as-SoR","type":"extras"},{"doc":"Under the read-through pattern, the cache is configured with a loader component that knows how to load data from the system-of-record (SoR). When the cache is asked for the value associated with a given key and such an entry does not exist within the cache, the cache invokes the loader to retrieve the value from the SoR, then caches the value, then returns it to the caller. The next time the cache is asked for the value for the same key it can be returned from the cache without using the loader (unless the entry has been evicted or expired). This pattern can be easily implemented using cache decorator as follows: defmodule MyApp.Example do use Nebulex.Caching.Decorators alias MyApp.Cache @decorate cache(cache: Cache, key: name) def get_by_name(name, age) do # your logic (the loader to retrieve the value from the SoR) end @decorate cache(cache: Cache, key: age, opts: [ttl: 3600]) def get_by_age(age) do # your logic (the loader to retrieve the value from the SoR) end @decorate cache(cache: Cache) def all(query) do # your logic (the loader to retrieve the value from the SoR) end end As you can see, the loader to retrieve the value from the system-of-record (SoR) is the function logic itself.","ref":"caching-decorators.html#read-through","title":"Cache Usage Patterns via Nebulex.Caching.Decorators - Read-through","type":"extras"},{"doc":"Under the write-through pattern, the cache is configured with a writer component that knows how to write data to the system-of-record (SoR). When the cache is asked to store a value for a key, the cache invokes the writer to store the value in the SoR, as well as updating (or deleting) the cache. This pattern can be implemented using defevict or defupdatable. When the data is written to the system-of-record (SoR), you can update the cached value associated with the given key using defupdatable, or just delete it using defevict. defmodule MyApp.Example do use Nebulex.Caching.Decorators alias MyApp.Cache # When the data is written to the SoR, it is updated in the cache @decorate update(cache: Cache, key: something) def update(something) do # Write data to the SoR (most likely the Database) end # When the data is written to the SoR, it is deleted (evicted) from the cache @decorate evict(cache: Cache, key: something) def update_something(something) do # Write data to the SoR (most likely the Database) end end As you can see, the logic to write data to the system-of-record (SoR) is the function logic itself.","ref":"caching-decorators.html#write-through","title":"Cache Usage Patterns via Nebulex.Caching.Decorators - Write-through","type":"extras"}]